use crate::bussy::{channels, Command, Message, PubSub};
use crate::utils::config::Config;
use crate::utils::error::Error;
use nostr::Event;
use std::sync::Arc;

pub struct Pipeline {
    pubsub: Arc<PubSub>,
    config: Arc<Config>,
}

impl Pipeline {
    pub fn new(pubsub: Arc<PubSub>, config: Arc<Config>) -> Self {
        Self { pubsub, config }
    }

    pub async fn start(&mut self) -> Result<(), Error> {
        let mut subscriber = self.pubsub.subscribe(channels::MSG_PIPELINE).await;

        while let Ok(message) = subscriber.recv().await {
            log::info!("[Pipeline] received message: {:?}", message);
            let channel;
            let command = match message.content {
                Command::PipelineReqEvent(client_id, event) => {
                    match self.handle_event(client_id, event).await {
                        Ok(_) => {
                            log::info!("[Pipeline] handle_event completed");
                            channel = message.source;
                            Command::Noop
                        }
                        Err(e) => {
                            channel = channels::MSG_RELAY;
                            Command::ServiceError(e)
                        }
                    }
                }
                Command::DbResOkWithStatus(client_id, message) => {
                    channel = channels::MSG_RELAY;
                    Command::PipelineResRelayMessageOk(client_id, message)
                }
                _ => {
                    channel = channels::MSG_RELAY;
                    Command::Noop
                }
            };
            if command != Command::Noop {
                let message = Message {
                    source: channels::MSG_PIPELINE,
                    content: command,
                };

                log::info!(
                    "[Pipeline] channel: {} - publishing new message: {:?}",
                    channel,
                    message
                );

                self.pubsub.publish(channel, message).await;
            }
        }

        Ok(())
    }

    pub async fn handle_event(
        &self,
        client_id: uuid::Uuid,
        event: Box<Event>,
    ) -> Result<(), Error> {
        let store_event_task = self.store_event(client_id, event.clone());
        let process_deletions_task = self.process_deletions(client_id, event.clone());
        let track_hashtags_task = self.track_hashtags(event.clone());
        let process_media_task = self.process_media(event.clone());
        let stream_out_task = self.stream_out(event.clone());
        let broadcast_task = self.broadcast(event.clone());

        let (
            store_event_result,
            process_deletions_result,
            track_hashtags_result,
            process_media_result,
            stream_out_result,
            broadcast_result,
        ) = tokio::join!(
            store_event_task,
            process_deletions_task,
            track_hashtags_task,
            process_media_task,
            stream_out_task,
            broadcast_task
        );

        match (
            store_event_result,
            process_deletions_result,
            track_hashtags_result,
            process_media_result,
            stream_out_result,
            broadcast_result,
        ) {
            (Ok(_), Ok(_), Ok(_), Ok(_), Ok(_), Ok(_)) => {
                log::info!("[Pipeline] Tasks finished successfully");
                Ok(())
            }
            _ => {
                log::error!("[Pipeline] One or more futures returned an error.");
                Err(Error::internal_with_message(
                    "[Pipeline] One or more futures returned an error.",
                ))
            }
        }
    }

    async fn store_event(&self, client_id: uuid::Uuid, event: Box<Event>) -> Result<(), Error> {
        log::info!("[Pipeline] store_event: processing...");
        if event.kind.as_u32() == 5 && event.pubkey == *self.config.get_admin_pubkey() {
            log::info!("[Pipeline] store_event: skipping admin event");
            return Ok(());
        }
        if event.kind.is_ephemeral() {
            log::info!("[Pipeline] store_event: skipping ephemeral event");
            return Ok(());
        }

        log::info!("[Pipeline] store_event: writing event to DB");
        self.pubsub
            .publish(
                channels::MSG_NOOSE,
                Message {
                    source: channels::MSG_PIPELINE,
                    content: Command::DbReqWriteEvent(client_id, event),
                },
            )
            .await;

        log::info!("[Pipeline] store_event: finished");
        Ok(())
    }

    async fn process_deletions(
        &self,
        client_id: uuid::Uuid,
        event: Box<Event>,
    ) -> Result<(), Error> {
        log::info!("[Pipeline] process_deletions: processing...");
        if event.kind.as_u32() == 5 && event.pubkey == *self.config.get_admin_pubkey() {
            self.pubsub
                .publish(
                    channels::MSG_NOOSE,
                    Message {
                        source: channels::MSG_PIPELINE,
                        content: Command::DbReqDeleteEvents(client_id, event),
                    },
                )
                .await;
        }

        log::info!("[Pipeline] process_deletions: finished");
        Ok(())
    }

    async fn track_hashtags(&self, event: Box<Event>) -> Result<(), Error> {
        log::info!("[Pipeline] track_hashtags: processing...");
        log::info!("[Pipeline] track_hashtags: finished");
        Ok(())
    }

    async fn process_media(&self, event: Box<Event>) -> Result<(), Error> {
        log::info!("[Pipeline] process_media: processing...");
        log::info!("[Pipeline] process_media: finished");
        Ok(())
    }

    async fn stream_out(&self, event: Box<Event>) -> Result<(), Error> {
        log::info!("[Pipeline] stream_out: processing...");
        let message = Message {
            source: channels::MSG_PIPELINE,
            content: Command::PipelineResStreamOutEvent(event),
        };
        self.pubsub.publish(channels::MSG_RELAY, message).await;

        log::info!("[Pipeline] stream_out: finished");
        Ok(())
    }

    async fn broadcast(&self, event: Box<Event>) -> Result<(), Error> {
        log::info!("[Pipeline] broadcast: processing...");
        log::info!("[Pipeline] broadcast: finished");
        Ok(())
    }
}
