use super::{
    db::Noose,
    migrations::MigrationRunner,
    user::{Nip05Profile, Nip05Table, UserRow},
};
use crate::{
    bussy::{channels, Command, Message, PubSub},
    usernames::dto::UserBody,
    utils::{config::Config as ServiceConfig, error::Error, structs::Subscription},
};
use async_trait::async_trait;
use deadpool_sqlite::{Config, Object, Pool, Runtime};
use nostr::{
    nips::nip01::Coordinate, secp256k1::XOnlyPublicKey, Event, EventId, Filter, JsonUtil,
    RelayMessage, TagKind, Timestamp, Url,
};
use nostr_database::{Backend, DatabaseOptions, NostrDatabase, Order};

use sea_query::{extension::sqlite::SqliteExpr, Order as SqOrder, Query, SqliteQueryBuilder};
use sea_query_rusqlite::RusqliteBinder;
use std::sync::Arc;
use std::{collections::HashSet, str::FromStr};

use crate::noose::sqlite_tables::*;

#[derive(Debug)]
pub struct NostrSqlite {
    pool: Pool,
}

impl NostrSqlite {
    pub async fn new(config: Arc<ServiceConfig>) -> Self {
        let cfg = Config::new(config.get_db_path());
        let pool = cfg.create_pool(Runtime::Tokio1).unwrap();

        if NostrSqlite::run_migrations(&pool).await {
            return Self { pool };
        }

        panic!("Pool's closed");
    }

    async fn run_migrations(pool: &Pool) -> bool {
        let connection = pool.get().await.unwrap();
        connection.interact(MigrationRunner::up).await.unwrap();
        connection
            .interact(|conn| {
                conn.pragma_update(None, "encoding", "UTF-8").unwrap();
                conn.pragma_update(None, "journal_mode", "WAL").unwrap();
                conn.pragma_update(None, "foreign_keys", "ON").unwrap();
                conn.pragma_update(None, "auto_vacuum", "FULL").unwrap();
                conn.pragma_update(None, "journal_size_limit", "32768")
                    .unwrap();
                conn.pragma_update(None, "mmap_size", "17179869184")
                    .unwrap();
            })
            .await
            .unwrap();

        true
    }

    async fn get_connection(&self) -> Result<Object, Error> {
        match self.pool.get().await {
            Ok(connection) => Ok(connection),
            Err(err) => Err(Error::internal(err.into())),
        }
    }

    async fn check_event_replaceble(&self, event: &Event) -> Result<bool, Error> {
        if event.is_replaceable() {
            let pubkey = event.pubkey.to_string();
            let kind = event.kind.as_u64();
            let created_at = event.created_at.as_i64();
            let Ok(connection) = self.get_connection().await else {
                return Err(Error::internal_with_message("Unable to get DB connection"));
            };

            let Ok(event_saved) = connection
                .interact(move |conn| -> Result<bool, Error> {
                    let tx = conn.transaction().unwrap();
                    log::debug!("new event is replaceable - searching for previously stored event");
                    let (sql, values) = Query::select()
                        .from(EventsTable::Table)
                        .columns([EventsTable::EventId])
                        .and_where(sea_query::Expr::col(EventsTable::Pubkey).eq(&pubkey))
                        .and_where(sea_query::Expr::col(EventsTable::Kind).eq(kind))
                        .and_where(sea_query::Expr::col(EventsTable::CreatedAt).gte(created_at))
                        .limit(1)
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Ok(res) = tx.execute(sql.as_str(), &*values.as_params()) {
                        if res > 0 {
                            log::debug!("deleting older replaceable event from events table");
                            let (sql, values) = Query::delete()
                                .from_table(EventsTable::Table)
                                .and_where(
                                    sea_query::Expr::col((EventsTable::Table, EventsTable::Kind))
                                        .eq(kind),
                                )
                                .and_where(
                                    sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey))
                                        .eq(&pubkey),
                                )
                                .and_where(
                                    sea_query::Expr::col((
                                        EventsTable::Table,
                                        EventsTable::EventId,
                                    ))
                                    .not_in_subquery(
                                        Query::select()
                                            .from(EventsTable::Table)
                                            .column(EventsTable::EventId)
                                            .and_where(
                                                sea_query::Expr::col(EventsTable::Kind).eq(kind),
                                            )
                                            .and_where(
                                                sea_query::Expr::col(EventsTable::Pubkey)
                                                    .eq(&pubkey),
                                            )
                                            .order_by(
                                                EventsTable::CreatedAt,
                                                sea_query::Order::Desc,
                                            )
                                            .limit(1)
                                            .to_owned(),
                                    ),
                                )
                                .build_rusqlite(SqliteQueryBuilder);

                            if let Ok(results) = tx.execute(sql.as_str(), &*values.as_params()) {
                                if results > 0 {
                                    log::info!(
                                "removed {} older replaceable kind {} events for author: {:?}",
                                results,
                                kind,
                                pubkey
                            );
                                }
                            }
                        }
                    };

                    tx.commit().unwrap();

                    Ok(true)
                })
                .await
            else {
                return Err(Error::internal_with_message("Failed to execute query"));
            };
        }
        Ok(true)
    }

    async fn check_event_parametrized_replaceble(&self, event: &Event) -> Result<bool, Error> {
        if event.is_parameterized_replaceable() {
            let identifier = match event.identifier() {
                Some(id) => id.to_string(),
                None => return Ok(true),
            };
            let pubkey = event.pubkey.to_string();
            let kind = event.kind.as_u64();
            let created_at = event.created_at.as_i64();

            let Ok(connection) = self.get_connection().await else {
                return Err(Error::internal_with_message("Unable to get DB connection"));
            };

            let Ok(event_saved) = connection
                .interact(move |conn| -> Result<bool, Error> {
                    let tx = conn.transaction().unwrap();
                    log::debug!(
                    "new event is parametrized replaceable - searching for previously stored event"
                    );

                    let (sql, values) = Query::select()
                        .from(EventsTable::Table)
                        .column((EventsTable::Table, EventsTable::EventId))
                        .left_join(
                            TagsTable::Table,
                            sea_query::Expr::col((TagsTable::Table, TagsTable::EventId))
                                .equals((EventsTable::Table, EventsTable::EventId)),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey))
                                .eq(&pubkey),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::Kind)).eq(kind),
                        )
                        .and_where(sea_query::Expr::col((TagsTable::Table, TagsTable::Tag)).eq("d"))
                        .and_where(
                            sea_query::Expr::col((TagsTable::Table, TagsTable::Value))
                                .eq(&identifier),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                                .gte(created_at),
                        )
                        .limit(1)
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Ok(results) = tx.execute(sql.as_str(), &*values.as_params()) {
                        if results > 0 {
                            log::info!(
                                "deleting older parametrized replaceable event from events table"
                            );
                            let (sql, values) = Query::delete()
                                .from_table(EventsTable::Table)
                                .and_where(
                                    sea_query::Expr::col((
                                        EventsTable::Table,
                                        EventsTable::EventId,
                                    ))
                                    .in_subquery(
                                        Query::select()
                                            .from(EventsTable::Table)
                                            .column((EventsTable::Table, EventsTable::EventId))
                                            .left_join(
                                                TagsTable::Table,
                                                sea_query::Expr::col((
                                                    TagsTable::Table,
                                                    TagsTable::EventId,
                                                ))
                                                .equals((EventsTable::Table, EventsTable::EventId)),
                                            )
                                            .and_where(
                                                sea_query::Expr::col((
                                                    EventsTable::Table,
                                                    EventsTable::Kind,
                                                ))
                                                .eq(kind),
                                            )
                                            .and_where(
                                                sea_query::Expr::col((
                                                    EventsTable::Table,
                                                    EventsTable::Pubkey,
                                                ))
                                                .eq(&pubkey),
                                            )
                                            .and_where(
                                                sea_query::Expr::col((
                                                    TagsTable::Table,
                                                    TagsTable::Tag,
                                                ))
                                                .eq("d"),
                                            )
                                            .and_where(
                                                sea_query::Expr::col((
                                                    TagsTable::Table,
                                                    TagsTable::Value,
                                                ))
                                                .eq(&identifier),
                                            )
                                            .to_owned(),
                                    ),
                                )
                                .build_rusqlite(SqliteQueryBuilder);

                            if let Ok(results) = tx.execute(sql.as_str(), &*values.as_params()) {
                                if results > 0 {
                                    log::info!("removed {} older parameterized replaceable kind {} events for author: {:?}", results, kind, &pubkey);
                                }
                            }
                        }
                    };

                    tx.commit().unwrap();
                    Ok(false)
                })
                .await
            else {
                return Err(Error::internal_with_message("Failed to execute query"));
            };
        };

        Ok(true)
    }

    async fn save_event(&self, event: &Event) -> Result<bool, Error> {
        let event = event.clone();

        let id = event.id.to_string();
        let kind = event.kind.as_u64();
        let pubkey = event.pubkey.to_string();
        let content = event.content.to_string();
        let created_at = event.created_at.as_i64();
        let tags = serde_json::to_string(&event.tags).unwrap();
        let sig = event.sig.to_string();

        let ids: Vec<String> = event.event_ids().map(|eid| eid.to_string()).collect();

        if self.event_is_too_old(&event) {
            return Ok(false);
        }

        if event.is_ephemeral() {
            return Ok(false);
        }

        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Ok(res) = self.check_event_replaceble(&event).await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Ok(res) = self.check_event_parametrized_replaceble(&event).await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Ok(event_saved) = connection
            .interact(move |conn| -> Result<bool, Error> {
                let tx = conn.transaction().unwrap();

                if event.kind == nostr::Kind::EventDeletion {
                    // Delete from Events
                    let (sql, values) = Query::delete()
                        .from_table(EventsTable::Table)
                        .and_where(sea_query::Expr::col(EventsTable::Kind).ne(5))
                        .and_where(sea_query::Expr::col(EventsTable::Pubkey).eq(&pubkey))
                        .and_where(sea_query::Expr::col(EventsTable::EventId).is_in(&ids))
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                        tx.rollback().unwrap();

                        return Ok(false);
                    }

                    // Delete from EventsFTS
                    let (sql, values) = Query::delete()
                        .from_table(EventsFTSTable::Table)
                        .and_where(sea_query::Expr::col(EventsFTSTable::EventId).is_in(ids))
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                        tx.rollback().unwrap();

                        return Ok(false);
                    }

                    match tx.commit() {
                        Ok(_) => return Ok(true),
                        Err(err) => {
                            log::error!("Error during transaction commit: {}", err);
                            return Ok(false);
                        }
                    }
                }

                if event.kind() == nostr::Kind::Metadata {
                    // Try to delete old profile first
                    let (sql, value) = Query::delete()
                        .from_table(ProfilesTable::Table)
                        .and_where(sea_query::Expr::col(ProfilesTable::Pubkey).eq(pubkey.clone()))
                        .build_rusqlite(SqliteQueryBuilder);
                    
                    if let Err(err) = tx.execute(sql.as_str(), &*value.as_params()) {
                        tx.rollback().unwrap();

                        log::debug!("Failed to delete old Profile record: {}", err);
                        return Ok(false);
                    }
                    
                    let (sql, value) = Query::insert()
                        .into_table(ProfilesTable::Table)
                        .columns([ProfilesTable::Pubkey])
                        .values_panic([pubkey.clone().into()])
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*value.as_params()) {
                        tx.rollback().unwrap();
                        
                        log::debug!("Failed to store Profile");
                        return Ok(false);
                    }

                    let Ok(metadata) = nostr::Metadata::from_json(content.clone()) else {
                        log::debug!("Failed to parse metadata");
                        return Err(Error::bad_request(
                            "Unable to parse metadata from 'content'",
                        ));
                    };

                    let metadata_custom_fields = serde_json::to_string(&metadata.custom);

                    let (sql, value) = Query::insert()
                        .into_table(MetadataTable::Table)
                        .columns([
                            MetadataTable::Pubkey,
                            MetadataTable::Name,
                            MetadataTable::DisplayName,
                            MetadataTable::About,
                            MetadataTable::Website,
                            MetadataTable::Picture,
                            MetadataTable::Banner,
                            MetadataTable::Nip05,
                            MetadataTable::Lud06,
                            MetadataTable::Lud16,
                            MetadataTable::Custom,
                        ])
                        .values_panic([
                            pubkey.clone().into(),
                            metadata.name.unwrap_or_default().into(),
                            metadata.display_name.unwrap_or_default().into(),
                            metadata.about.unwrap_or_default().into(),
                            metadata.website.unwrap_or_default().into(),
                            metadata.picture.unwrap_or_default().into(),
                            metadata.banner.unwrap_or_default().into(),
                            metadata.nip05.unwrap_or_default().into(),
                            metadata.lud06.unwrap_or_default().into(),
                            metadata.lud16.unwrap_or_default().into(),
                            metadata_custom_fields.unwrap_or_default().into(),
                        ])
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*value.as_params()) {
                        tx.rollback().unwrap();

                        log::debug!("Failed to store Metadata: {}", err);
                        return Ok(false);
                    }
                }
                log::debug!("inserting new event in events");
                // Insert into Events table
                let (sql, values) = Query::insert()
                    .into_table(EventsTable::Table)
                    .columns([
                        EventsTable::EventId,
                        EventsTable::Content,
                        EventsTable::Kind,
                        EventsTable::Pubkey,
                        EventsTable::CreatedAt,
                        EventsTable::Tags,
                        EventsTable::Sig,
                    ])
                    .values_panic([
                        id.clone().into(),
                        content.clone().into(),
                        kind.into(),
                        pubkey.into(),
                        created_at.into(),
                        tags.into(),
                        sig.into(),
                    ])
                    .build_rusqlite(SqliteQueryBuilder);

                if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                    log::error!("Error inserting event into 'events' table: {}", err);
                    tx.rollback().unwrap();

                    return Ok(false);
                };

                // Insert into EventsFTS table
                log::debug!("inserting new event into eventsFTS");
                let (sql, values) = Query::insert()
                    .into_table(EventsFTSTable::Table)
                    .columns([EventsFTSTable::EventId, EventsFTSTable::Content])
                    .values_panic([id.clone().into(), content.into()])
                    .build_rusqlite(SqliteQueryBuilder);

                if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                    log::error!("Error inserting event into 'eventsFTS' table: {}", err);
                    tx.rollback().unwrap();

                    return Ok(false);
                }

                // Insert into Tags table
                log::debug!("inserting new event into tags");
                for tag in event.tags.clone() {
                    if Self::tag_is_indexable(&tag) {
                        let tag = tag.to_vec();
                        if tag.len() >= 2 {
                            let tag_name = &tag[0];
                            let tag_value = &tag[1];
                            if tag_name.len() == 1 {
                                let (sql, values) = Query::insert()
                                    .into_table(TagsTable::Table)
                                    .columns([TagsTable::Tag, TagsTable::Value, TagsTable::EventId])
                                    .values_panic([
                                        tag_name.into(),
                                        tag_value.into(),
                                        id.clone().into(),
                                    ])
                                    .build_rusqlite(SqliteQueryBuilder);

                                if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                                    log::error!("Error inserting event into 'tags' table: {}", err);
                                    tx.rollback().unwrap();

                                    return Ok(false);
                                }
                            }
                        }
                    }
                }

                match tx.commit() {
                    Ok(_) => Ok(true),
                    Err(err) => {
                        log::error!("Error during transaction commit: {}", err);
                        Ok(false)
                    }
                }
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        event_saved
    }

    fn tag_is_indexable(tag: &nostr::Tag) -> bool {
        matches!(
            tag.kind(),
            TagKind::E
                | TagKind::P
                | TagKind::UpperP
                | TagKind::A
                | TagKind::D
                | TagKind::G
                | TagKind::I
                | TagKind::M
                | TagKind::R
                | TagKind::T
        )
    }

    async fn has_event_already_been_saved(&self, event_id: &EventId) -> Result<bool, Error> {
        match self.get_event_by_id(*event_id).await {
            Ok(_) => Ok(true),
            Err(_) => Ok(false),
        }
    }

    async fn has_event_id_been_deleted(&self, event_id: &EventId) -> Result<bool, Error> {
        match self.get_event_by_id(*event_id).await {
            Ok(_) => Ok(false),
            Err(_) => Ok(true),
        }
    }

    /// Skip events that are older than 10 minutes
    fn event_is_too_old(&self, event: &Event) -> bool {
        if chrono::Utc::now().timestamp() - 600 > event.created_at.as_i64() {
            return true;
        }
        false
    }

    async fn set_event_id_seen(&self, event_id: EventId, relay_url: Url) -> Result<(), Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let event_id = event_id.to_string();
        let relay_url = relay_url.to_string();

        let Ok(query_result) = connection
            .interact(|conn| -> Result<usize, Error> {
                let (sql, values) = Query::insert()
                    .into_table(EventSeenByRelaysTable::Table)
                    .columns([
                        EventSeenByRelaysTable::EventId,
                        EventSeenByRelaysTable::RelayURL,
                    ])
                    .values_panic([event_id.into(), relay_url.into()])
                    .build_rusqlite(SqliteQueryBuilder);

                match conn.execute(sql.as_str(), &*values.as_params()) {
                    Ok(updated_rows) => Ok(updated_rows),
                    Err(err) => Err(Error::internal(err.into())),
                }
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        if let Ok(event_id_inserted) = query_result {
            if event_id_inserted > 0 {
                return Ok(());
            }

            Err(Error::internal_with_message("Failed to insert new record"))
        } else {
            Err(Error::internal_with_message("Failed to insert new record"))
        }
    }

    async fn get_event_seen_on_relays(
        &self,
        event_id: EventId,
    ) -> Result<Option<HashSet<Url>>, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };
        let event_id = event_id.to_string();

        let Ok(query_result) = connection
            .interact(|conn| -> Result<Option<HashSet<Url>>, Error> {
                let (sql, values) = Query::select()
                    .from(EventSeenByRelaysTable::Table)
                    .column(EventSeenByRelaysTable::RelayURL)
                    .and_where(sea_query::Expr::col(EventSeenByRelaysTable::EventId).eq(event_id))
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let mut urls: HashSet<Url> = HashSet::new();
                while let Ok(Some(record)) = rows.next() {
                    if let Ok(url_string) = record.get::<_, String>("relay_url") {
                        if let Ok(relay_url) = Url::from_str(&url_string) {
                            urls.insert(relay_url);
                        }
                    }
                }

                if !urls.is_empty() {
                    return Ok(Some(urls));
                }

                Ok(None)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        let Ok(found_relay_urls) = query_result else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        Ok(found_relay_urls)
    }

    async fn has_event_already_been_seen(&self, event_id: &EventId) -> Result<bool, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let event_id = event_id.to_string();

        let Ok(query_result) = connection
            .interact(|conn| -> Result<bool, Error> {
                let (sql, values) = Query::select()
                    .expr_as(
                        sea_query::Expr::exists(
                            Query::select()
                                .column(EventSeenByRelaysTable::EventId)
                                .from(EventSeenByRelaysTable::Table)
                                .and_where(
                                    sea_query::Expr::col(EventSeenByRelaysTable::EventId)
                                        .eq(event_id),
                                )
                                .limit(1)
                                .take(),
                        ),
                        sea_query::Alias::new("event_exists"),
                    )
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let exists = match rows.next() {
                    Ok(_) => 1,
                    Err(_) => 0,
                };

                Ok(exists == 1)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        if let Ok(result) = query_result {
            return Ok(result);
        }

        Ok(false)
    }

    async fn admin_delete_events(&self, event: &Event) -> Result<RelayMessage, Error> {
        let event_ids: Vec<String> = event.event_ids().map(|e| e.to_string()).collect();
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };
        let Ok(query_result) = connection
            .interact(
                |conn: &mut rusqlite::Connection| -> Result<RelayMessage, Error> {
                    let (sql, values) = Query::delete()
                        .from_table(EventsTable::Table)
                        .and_where(sea_query::Expr::col(EventsTable::EventId).is_in(event_ids))
                        .build_rusqlite(SqliteQueryBuilder);

                    match conn.execute(sql.as_str(), &*values.as_params()) {
                        Ok(affected_rows) => {
                            let message = format!("{} events deleted", affected_rows);
                            Ok(RelayMessage::notice(message))
                        }
                        Err(err) => {
                            log::error!("[admin_delete_events] Failed to execute query: {}", err);

                            Ok(RelayMessage::notice("unable to delete events"))
                        }
                    }
                },
            )
            .await
        else {
            log::error!("[admin_delete_events] Failed to execute query");
            return Err(Error::internal_with_message(
                "Failed to execute query 'admin_delete_events'",
            ));
        };

        let Ok(relay_message) = query_result else {
            let message = RelayMessage::Notice {
                message: "unable to acquire pool connection".to_string(),
            };
            return Ok(message);
        };

        Ok(relay_message)
    }

    async fn get_event_by_id(&self, event_id: EventId) -> Result<Event, Error> {
        let eid = event_id.to_string();
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };
        let Ok(query_result) = connection
            .interact(|conn: &mut rusqlite::Connection| -> Result<Event, Error> {
                let (sql, value) = Query::select()
                    .from(EventsTable::Table)
                    .columns([
                        EventsTable::EventId,
                        EventsTable::Content,
                        EventsTable::Kind,
                        EventsTable::Pubkey,
                        EventsTable::CreatedAt,
                        EventsTable::Tags,
                        EventsTable::Sig,
                    ])
                    .and_where(sea_query::Expr::col(EventsTable::EventId).eq(eid))
                    .limit(1)
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let row = stmt
                    .query_row::<EventRow, _, _>(&*value.as_params(), |row| Ok(EventRow::from(row)))
                    .unwrap();

                let event = row.clone().to_event();

                if event.is_expired() {
                    return Err(Error::internal_with_message(
                        "Event has expired. Ignoring...",
                    ));
                }

                Ok(event)
            })
            .await
        else {
            return Err(Error::internal_with_message(
                "Failed to execute query 'get_event_by_id'",
            ));
        };

        let Ok(event) = query_result else {
            return Err(Error::internal_with_message(
                "Failed to get event from row in 'get_event_by_id'",
            ));
        };

        Ok(event)
    }

    async fn wipe(&self) -> Result<(), Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        match connection
            .interact(|conn| -> Result<(), Error> {
                let Ok(result) = conn.set_db_config(
                    rusqlite::config::DbConfig::SQLITE_DBCONFIG_RESET_DATABASE,
                    true,
                ) else {
                    return Err(Error::internal_with_message(
                        "Failed to set SQLITE_DBCONFIG_RESET_DATABASE to true",
                    ));
                };
                let Ok(result) = conn.execute("VACUUM;", []) else {
                    return Err(Error::internal_with_message("Failed to run VACUUM command"));
                };
                let Ok(result) = conn.set_db_config(
                    rusqlite::config::DbConfig::SQLITE_DBCONFIG_RESET_DATABASE,
                    false,
                ) else {
                    return Err(Error::internal_with_message(
                        "Failed to set SQLITE_DBCONFIG_RESET_DATABASE to false",
                    ));
                };

                Ok(())
            })
            .await
        {
            Ok(_) => {
                if NostrSqlite::run_migrations(&self.pool).await {
                    return Ok(());
                }

                Err(Error::internal_with_message("Failed to run migrations"))
            }
            Err(err) => Err(Error::internal_with_message(err.to_string())),
        }
    }

    fn get_filter_query(&self, filter: &nostr::Filter) -> sea_query::SelectStatement {
        let mut query = Query::select()
            .column((EventsTable::Table, EventsTable::EventId))
            .column((EventsTable::Table, EventsTable::Content))
            .columns([
                EventsTable::Kind,
                EventsTable::Pubkey,
                EventsTable::CreatedAt,
                EventsTable::Tags,
                EventsTable::Sig,
            ])
            .from(EventsTable::Table)
            .order_by(EventsTable::CreatedAt, sea_query::Order::Desc)
            .to_owned();

        if !filter.ids.is_empty() {
            let ids = filter.ids.iter().map(|id| id.to_string());
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::EventId)).is_in(ids),
                )
                .to_owned();
        }

        if !filter.kinds.is_empty() {
            let kinds: Vec<u32> = filter.kinds.iter().map(|kind| kind.as_u32()).collect();
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::Kind)).is_in(kinds),
                )
                .to_owned();
        }

        if !filter.authors.is_empty() {
            let authors = filter.authors.iter().map(|author| author.to_string());
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey)).is_in(authors),
                )
                .to_owned();
        }

        if let Some(since) = filter.since {
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                        .gte(since.as_u64()),
                )
                .to_owned();
        }

        if let Some(until) = filter.until {
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                        .lte(until.as_u64()),
                )
                .to_owned();
        }

        if let Some(limit) = filter.limit {
            query = query.limit(limit as u64).to_owned();
        }

        if !filter.generic_tags.is_empty() {
            query = query
                .left_join(
                    TagsTable::Table,
                    sea_query::Expr::col((TagsTable::Table, TagsTable::EventId))
                        .equals((EventsTable::Table, EventsTable::EventId)),
                )
                .to_owned();

            filter.generic_tags.iter().for_each(|(tag, values)| {
                let values = values.iter().map(|val| val.to_string());
                query = query
                    .and_where(
                        sea_query::Expr::col((TagsTable::Table, TagsTable::Tag))
                            .eq(tag.to_string()),
                    )
                    .and_where(
                        sea_query::Expr::col((TagsTable::Table, TagsTable::Value)).is_in(values),
                    )
                    .to_owned();
            });
        }

        if let Some(search) = &filter.search {
            query = query
                .inner_join(
                    EventsFTSTable::Table,
                    sea_query::Expr::col((EventsTable::Table, EventsTable::EventId))
                        .equals((EventsFTSTable::Table, EventsFTSTable::EventId)),
                )
                .and_where(
                    sea_query::Expr::col((EventsFTSTable::Table, EventsFTSTable::Content))
                        .matches(search),
                )
                .to_owned();
        }

        query
    }

    fn get_filters_query(&self, filters: Vec<Filter>) -> Option<sea_query::SelectStatement> {
        filters
            .iter()
            .map(|filter| {
                Query::select()
                    .column((EventsTable::Table, EventsTable::EventId))
                    .column((EventsTable::Table, EventsTable::Content))
                    .columns([
                        EventsTable::Kind,
                        EventsTable::Pubkey,
                        EventsTable::CreatedAt,
                        EventsTable::Tags,
                        EventsTable::Sig,
                    ])
                    .from_subquery(
                        self.get_filter_query(filter),
                        sea_query::Alias::new("events"),
                    )
                    .to_owned()
            })
            .reduce(|mut result, query| result.union(sea_query::UnionType::All, query).to_owned())
    }

    async fn query(&self, filters: Vec<Filter>, order: Order) -> Result<Vec<Event>, Error> {
        log::debug!("making query from filters...");

        // let event_ids: Vec<String> = event.event_ids().map(|e| e.to_string()).collect();
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let sq_order = match order {
            Order::Asc => SqOrder::Asc,
            Order::Desc => SqOrder::Desc,
        };
        let Some(mut sql_statement) = self.get_filters_query(filters) else {
            return Err(Error::internal_with_message("Failed to build SQL Query"));
        };

        let Ok(query_result) = connection
            .interact(move |conn| -> Result<Vec<Event>, Error> {
                let (sql, values) = sql_statement
                    .order_by(EventsTable::CreatedAt, sq_order.to_owned())
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let mut event_vec: Vec<Event> = vec![];
                while let Ok(Some(row)) = rows.next() {
                    let event = EventRow::from(row).to_event();
                    if !event.is_expired() {
                        event_vec.push(event);
                    }
                }

                Ok(event_vec)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        query_result
    }

    async fn count(&self, filters: Vec<Filter>) -> Result<usize, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Some(mut sql_statement) = self.get_filters_query(filters) else {
            return Err(Error::internal_with_message("Failed to build SQL Query"));
        };

        let Ok(query_result) = connection
            .interact(move |conn| {
                let (sql, values) = sql_statement
                    .clear_selects()
                    .expr_as(
                        sea_query::Func::count(sea_query::Expr::col((
                            EventsTable::Table,
                            EventsTable::EventId,
                        ))),
                        sea_query::Alias::new("count"),
                    )
                    .build_rusqlite(SqliteQueryBuilder);

                if let Ok(result) = conn.query_row(sql.as_str(), &*values.as_params(), |row| {
                    let count: usize = row.get(0).unwrap();
                    Ok(count)
                }) {
                    return Ok(result);
                }

                Err(Error::internal_with_message(
                    "Failed to get event counts by filters",
                ))
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        query_result
    }

    async fn event_ids_by_filters(
        &self,
        filters: Vec<Filter>,
        order: Order,
    ) -> Result<Vec<EventId>, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let sq_order = match order {
            Order::Asc => SqOrder::Asc,
            Order::Desc => SqOrder::Desc,
        };

        let Some(mut sql_statement) = self.get_filters_query(filters) else {
            return Err(Error::internal_with_message("Failed to build SQL Query"));
        };

        let Ok(query_result) = connection
            .interact(move |conn| {
                let (sql, values) = sql_statement
                    .order_by(EventsTable::CreatedAt, sq_order.to_owned())
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let mut event_vec: Vec<EventId> = vec![];
                while let Ok(Some(row)) = rows.next() {
                    let event = EventRow::from(row).to_event();
                    if !event.is_expired() {
                        event_vec.push(event.id);
                    }
                }

                Ok(event_vec)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        query_result
    }

    async fn has_coordinate_been_deleted(
        &self,
        coordinate: &Coordinate,
        timestamp: Timestamp,
    ) -> Result<bool, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let ident = if coordinate.identifier.is_empty() {
            format!("{}:{}", coordinate.kind, coordinate.pubkey)
        } else {
            format!(
                "{}:{}:{}",
                coordinate.kind, coordinate.pubkey, coordinate.identifier
            )
        };

        let Ok(query_result) = connection
            .interact(
                move |conn: &mut rusqlite::Connection| -> Result<bool, Error> {
                    let (sql, value) = Query::select()
                        .from(EventsTable::Table)
                        .columns([EventsTable::EventId, EventsTable::CreatedAt])
                        .left_join(
                            TagsTable::Table,
                            sea_query::Expr::col((TagsTable::Table, TagsTable::EventId))
                                .equals((EventsTable::Table, EventsTable::EventId)),
                        )
                        .and_where(sea_query::Expr::col((TagsTable::Table, TagsTable::Tag)).eq("a"))
                        .and_where(
                            sea_query::Expr::col((TagsTable::Table, TagsTable::Value)).eq(ident),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                                .gte(timestamp.as_i64()),
                        )
                        .limit(1)
                        .build_rusqlite(SqliteQueryBuilder);

                    let mut stmt = conn.prepare(sql.as_str()).unwrap();
                    let mut rows = stmt.query(&*value.as_params()).unwrap();

                    if let Ok(Some(record)) = rows.next() {
                        return Ok(false);
                    }

                    Ok(true)
                },
            )
            .await
        else {
            return Err(Error::internal_with_message(
                "Failed to execute query 'get_event_by_id'",
            ));
        };

        query_result
    }

    async fn get_nip05_profile(&self, username: String) -> Result<Nip05Profile, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Ok(query_result) = connection
            .interact(
                move |conn: &mut rusqlite::Connection| -> Result<Nip05Profile, Error> {
                    let (sql, value) = Query::select()
                        .from(Nip05Table::Table)
                        .columns([
                            Nip05Table::Pubkey,
                            Nip05Table::Username,
                            Nip05Table::Relays,
                            Nip05Table::JoinedAt,
                        ])
                        .and_where(sea_query::Expr::col(Nip05Table::Username).eq(&username))
                        .limit(1)
                        .build_rusqlite(SqliteQueryBuilder);

                    let Ok(res) = conn.query_row(sql.as_str(), &*value.as_params(), |row| {
                        let nip05_row: UserRow = row.into();
                        let nip05 = Nip05Profile::from(nip05_row);

                        Ok(nip05)
                    }) else {
                        return Err(Error::not_found("user", username));
                    };

                    Ok(res)
                },
            )
            .await
        else {
            return Err(Error::internal_with_message(
                "Failed to execute query 'get_nip05_profile'",
            ));
        };

        query_result
    }

    async fn get_profile(
        &self,
        public_key: XOnlyPublicKey,
    ) -> Result<nostr_database::Profile, Error> {
        let pk = public_key.to_string();

        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Ok(query_result) = connection
            .interact(
                move |conn: &mut rusqlite::Connection| -> Result<nostr_database::Profile, Error> {
                    let (sql, value) = Query::select()
                        .from(MetadataTable::Table)
                        .columns([
                            MetadataTable::Pubkey,
                            MetadataTable::Name,
                            MetadataTable::DisplayName,
                            MetadataTable::About,
                            MetadataTable::Website,
                            MetadataTable::Picture,
                            MetadataTable::Banner,
                            MetadataTable::Nip05,
                            MetadataTable::Lud06,
                            MetadataTable::Lud16,
                            MetadataTable::Custom,
                        ])
                        .and_where(sea_query::Expr::col(MetadataTable::Pubkey).eq(&pk))
                        .limit(1)
                        .build_rusqlite(SqliteQueryBuilder);

                    let Ok(res) = conn.query_row(sql.as_str(), &*value.as_params(), |row| {
                        let profile_row: ProfileRow = row.into();
                        let profile = nostr_database::Profile::from(&profile_row);

                        Ok(profile)
                    }) else {
                        return Err(Error::not_found("user", pk));
                    };

                    Ok(res)
                },
            )
            .await
        else {
            return Err(Error::internal_with_message(
                "Failed to execute query 'get_profile'",
            ));
        };

        query_result
    }

    async fn create_nip05_profile(&self, user: UserBody) -> Result<(), Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Ok(query_result) = connection
            .interact(
                move |conn: &mut rusqlite::Connection| -> Result<(), Error> {
                    let (sql, values) = Query::insert()
                        .into_table(Nip05Table::Table)
                        .columns([
                            Nip05Table::Username,
                            Nip05Table::Pubkey,
                            Nip05Table::Relays,
                            Nip05Table::JoinedAt,
                        ])
                        .values_panic([
                            user.name.clone().into(),
                            user.get_pubkey().into(),
                            serde_json::to_string(&user.relays).unwrap().into(),
                            nostr::Timestamp::now().as_i64().into(),
                        ])
                        .build_rusqlite(SqliteQueryBuilder);

                    match conn.execute(sql.as_str(), &*values.as_params()) {
                        Ok(_) => Ok(()),
                        Err(e) => {
                            if e.to_string().contains("nip05.pubkey") {
                                return Err(Error::invalid_request_body("pubkey already taken"));
                            }
                            Err(Error::invalid_request_body("name already taken"))
                        }
                    }
                },
            )
            .await
        else {
            return Err(Error::internal_with_message(
                "Failed to execute query 'get_nip05_profile'",
            ));
        };

        query_result
    }
}

impl From<nostr_database::DatabaseError> for Error {
    fn from(value: nostr_database::DatabaseError) -> Self {
        Error::internal_with_message(value.to_string())
    }
}

impl From<Error> for nostr_database::DatabaseError {
    fn from(val: Error) -> Self {
        nostr_database::DatabaseError::backend(val)
    }
}

#[async_trait]
impl NostrDatabase for NostrSqlite {
    type Err = Error;

    /// Name of the backend database used (ex. lmdb, sqlite, indexeddb, ...)
    fn backend(&self) -> Backend {
        Backend::SQLite
    }

    /// Database options
    fn opts(&self) -> DatabaseOptions {
        DatabaseOptions { events: true }
    }

    /// Save [`Event`] into store
    ///
    /// Return `true` if event was successfully saved into database.
    ///
    /// **This method assume that [`Event`] was already verified**
    async fn save_event(&self, event: &Event) -> Result<bool, Self::Err> {
        self.save_event(event).await
    }

    /// Check if [`Event`] has already been saved
    async fn has_event_already_been_saved(&self, event_id: &EventId) -> Result<bool, Self::Err> {
        self.has_event_already_been_saved(event_id).await
    }

    /// Check if [`EventId`] has already been seen
    async fn has_event_already_been_seen(&self, event_id: &EventId) -> Result<bool, Self::Err> {
        self.has_event_already_been_seen(event_id).await
    }

    /// Check if [`EventId`] has been deleted
    async fn has_event_id_been_deleted(&self, event_id: &EventId) -> Result<bool, Self::Err> {
        self.has_event_id_been_deleted(event_id).await
    }

    /// Check if event with [`Coordinate`] has been deleted before [`Timestamp`]
    async fn has_coordinate_been_deleted(
        &self,
        coordinate: &Coordinate,
        timestamp: Timestamp,
    ) -> Result<bool, Self::Err> {
        self.has_coordinate_been_deleted(coordinate, timestamp)
            .await
    }

    /// Set [`EventId`] as seen by relay
    ///
    /// Useful for NIP65 (aka gossip)
    async fn event_id_seen(&self, event_id: EventId, relay_url: Url) -> Result<(), Self::Err> {
        self.set_event_id_seen(event_id, relay_url).await
    }

    /// Get list of relays that have seen the [`EventId`]
    async fn event_seen_on_relays(
        &self,
        event_id: EventId,
    ) -> Result<Option<HashSet<Url>>, Self::Err> {
        self.get_event_seen_on_relays(event_id).await
    }

    /// Get [`Event`] by [`EventId`]
    async fn event_by_id(&self, event_id: EventId) -> Result<Event, Self::Err> {
        self.get_event_by_id(event_id).await
    }

    /// Count number of [`Event`] found by filters
    ///
    /// Use `Filter::new()` or `Filter::default()` to count all events.
    async fn count(&self, filters: Vec<Filter>) -> Result<usize, Self::Err> {
        self.count(filters).await
    }

    /// Query store with filters
    async fn query(&self, filters: Vec<Filter>, order: Order) -> Result<Vec<Event>, Self::Err> {
        self.query(filters, order).await
    }

    /// Get event IDs by filters
    async fn event_ids_by_filters(
        &self,
        filters: Vec<Filter>,
        order: Order,
    ) -> Result<Vec<EventId>, Self::Err> {
        self.event_ids_by_filters(filters, order).await
    }

    /// Get `negentropy` items
    async fn negentropy_items(
        &self,
        filter: Filter,
    ) -> Result<Vec<(EventId, Timestamp)>, Self::Err> {
        Err(Error::internal_with_message(
            "negentropy is not currently supported",
        ))
    }

    /// Wipe all data
    async fn wipe(&self) -> Result<(), Self::Err> {
        self.wipe().await
    }
}

impl Noose for NostrSqlite {
    async fn start(&mut self, pubsub: Arc<PubSub>) -> Result<(), Error> {
        let mut subscriber = pubsub.subscribe(channels::MSG_NOOSE).await;

        while let Ok(message) = subscriber.recv().await {
            log::info!("[Noose] received message: {:?}", message);
            let command = match message.content {
                // Relay Events
                Command::DbReqWriteEvent(client_id, event) => match self.write_event(event).await {
                    Ok(status) => Command::DbResOkWithStatus(client_id, status),
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqFindEvent(client_id, subscriptioin) => {
                    match self.find_event(subscriptioin).await {
                        Ok(relay_messages) => {
                            Command::DbResRelayMessages(client_id, relay_messages)
                        }
                        Err(e) => Command::ServiceError(e),
                    }
                }
                Command::DbReqDeleteEvents(client_id, event) => {
                    match self.admin_delete_events(&event).await {
                        Ok(status) => Command::DbResOkWithStatus(client_id, status),
                        Err(e) => Command::ServiceError(e),
                    }
                }
                Command::DbReqEventCounts(client_id, subscriptioin) => {
                    match self.counts(subscriptioin).await {
                        Ok(relay_message) => Command::DbResEventCounts(client_id, relay_message),
                        Err(e) => Command::ServiceError(e),
                    }
                }
                // NIP-05
                Command::DbReqGetNIP05(username) => match self.get_nip05(username).await {
                    Ok(user) => Command::DbResNIP05(user),
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqCreateNIP05(user) => match self.create_nip05(user).await {
                    Ok(_) => Command::DbResOk,
                    Err(e) => Command::ServiceError(e),
                },
                // NIP-42
                Command::DbReqGetProfile(pubkey) => match self.profile(pubkey).await {
                    Ok(profile) => Command::DbResGetProfile(profile),
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqAddContact(profile_pub_key, contact_pub_key) => {
                    match self.add_contact(profile_pub_key, contact_pub_key).await {
                        Ok(_) => Command::DbResOk,
                        Err(e) => Command::ServiceError(e),
                    }
                }
                Command::DbReqRemoveContact(profile_pub_key, contact_pub_key) => {
                    match self.remove_contact(profile_pub_key, contact_pub_key).await {
                        Ok(_) => Command::DbResOk,
                        Err(e) => Command::ServiceError(e),
                    }
                }
                Command::DbReqGetContactPubkeys(profile_pub_key) => {
                    match self.contacts_public_keys(profile_pub_key).await {
                        Ok(contact_pubkeys) => Command::DbResGetContactPubkeys(contact_pubkeys),
                        Err(e) => Command::ServiceError(e),
                    }
                }
                Command::DbReqGetContacts(profile_pub_key) => {
                    match self.contacts(profile_pub_key).await {
                        Ok(contacts) => Command::DbResGetContacts(contacts),
                        Err(e) => Command::ServiceError(e),
                    }
                }
                _ => Command::Noop,
            };
            if command != Command::Noop {
                let channel = message.source;
                let message = Message {
                    source: channels::MSG_NOOSE,
                    content: command,
                };

                log::info!(
                    "[Noose] publishing new message: {:?} to channel {}",
                    message,
                    channel
                );
                
                pubsub.publish(channel, message).await;
            }
        }

        Ok(())
    }

    async fn write_event(&self, event: Box<nostr::Event>) -> Result<nostr::RelayMessage, Error> {
        match self.save_event(&event).await {
            Ok(status) => {
                let relay_message = nostr::RelayMessage::ok(event.id, status, "");
                Ok(relay_message)
            }
            Err(err) => Err(Error::bad_request(err.to_string())),
        }
    }

    async fn find_event(
        &self,
        subscription: Subscription,
    ) -> Result<Vec<nostr::RelayMessage>, Error> {
        match self.query(subscription.filters, Order::Desc).await {
            Ok(events) => {
                let relay_messages: Vec<RelayMessage> = events
                    .into_iter()
                    .map(|event| nostr::RelayMessage::event(subscription.id.clone(), event))
                    .collect();

                if relay_messages.is_empty() {
                    let eose = RelayMessage::eose(subscription.id.clone());
                    return Ok(vec![eose]);
                }

                Ok(relay_messages)
            }
            Err(err) => Err(Error::bad_request(err.to_string())),
        }
    }

    async fn counts(&self, subscription: Subscription) -> Result<nostr::RelayMessage, Error> {
        match self.count(subscription.filters).await {
            Ok(counts) => {
                let relay_message = nostr::RelayMessage::count(subscription.id, counts);
                Ok(relay_message)
            }
            Err(err) => Err(Error::internal_with_message(err.to_string())),
        }
    }

    async fn get_nip05(&self, username: String) -> Result<Nip05Profile, Error> {
        self.get_nip05_profile(username).await
    }

    async fn create_nip05(&self, user: UserBody) -> Result<(), Error> {
        self.create_nip05_profile(user).await
    }

    async fn profile(
        &self,
        public_key: nostr::prelude::XOnlyPublicKey,
    ) -> Result<nostr_database::Profile, Error> {
        self.get_profile(public_key).await
    }

    async fn add_contact(
        &self,
        profile_public_key: nostr::prelude::XOnlyPublicKey,
        contact_public_key: nostr::prelude::XOnlyPublicKey,
    ) -> Result<(), Error> {
        todo!()
    }

    async fn remove_contact(
        &self,
        profile_public_key: nostr::prelude::XOnlyPublicKey,
        contact_public_key: nostr::prelude::XOnlyPublicKey,
    ) -> Result<(), Error> {
        todo!()
    }

    async fn contacts_public_keys(
        &self,
        public_key: nostr::prelude::XOnlyPublicKey,
    ) -> Result<Vec<nostr::prelude::XOnlyPublicKey>, Error> {
        todo!()
    }

    async fn contacts(
        &self,
        public_key: nostr::prelude::XOnlyPublicKey,
    ) -> Result<std::collections::BTreeSet<nostr_database::Profile>, Error> {
        todo!()
    }
}

#[cfg(test)]
mod tests {
    use crate::noose::sqlite::*;
    use nostr::key::FromSkStr;
    use nostr::EventBuilder;

    #[tokio::test]
    async fn create_db() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        assert_eq!(db.pool.status().max_size, 64);
    }

    #[tokio::test]
    async fn event_row_to_event() {
        let event_id =
            "076b27df50401f2af598cb727281e1e36401592570052c0fb37543f50aba5b69".to_string();
        let event_row = EventRow {
            id: event_id.clone(),
            pubkey: "da9cfc5b7644aae5b3e7df1d224ef4d4a206d66d6e3b5c99a8aeda3b642c4cef".to_string(),
            created_at: 1706020262,
            kind: 1,
            tags: "[]".to_string(),
            sig: "7f92a6ebdeae4a9bd7a6e19f57b49e2c499fece644c63c6ff5776bdbafec5c524cfad1eecedef11a8701df2a826453914d39a897e0f8276a6ffc40e266cae059".to_string(),
            content: "hoh".to_string(),
        };

        let event = event_row.to_event();

        assert_eq!(event.id.to_string(), event_id);
    }

    #[tokio::test]
    async fn get_event_by_id() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert event
        let result = db.save_event(&event).await.unwrap();
        assert!(result);

        // Get event by id
        let eid = event.id;
        let result = db.get_event_by_id(eid).await.unwrap();

        assert_eq!(result.id, eid);
    }

    #[tokio::test]
    async fn event_seen_by_other_relays() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello to Other relay", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert event
        let result = db.save_event(&event).await.unwrap();
        assert!(result);

        // Set seen on other relays
        let url_0 = Url::from_str("wss://relay.zhitno.st").unwrap();
        db.set_event_id_seen(event.id, url_0).await.unwrap();

        let url_1 = Url::from_str("wss://relay.damus.io").unwrap();
        db.set_event_id_seen(event.id, url_1).await.unwrap();

        // Get status of seen event
        let result = db.get_event_seen_on_relays(event.id).await.unwrap();

        // Get event by id
        let eid = event.id;
        let result = db.get_event_by_id(eid).await.unwrap();

        assert_eq!(result.id, eid);
    }

    #[tokio::test]
    async fn has_event_already_been_seen() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello to Other relay", vec![])
            .to_event(&keys)
            .unwrap();

        // Set seen on other relays
        let url_0 = Url::from_str("wss://relay.zhitno.st").unwrap();
        db.set_event_id_seen(event.id, url_0).await.unwrap();

        let url_1 = Url::from_str("wss://relay.damus.io").unwrap();
        db.set_event_id_seen(event.id, url_1).await.unwrap();

        // Get status of seen event
        let result = db.has_event_already_been_seen(&event.id).await.unwrap();

        assert!(result);
    }

    #[tokio::test]
    async fn wipe_db() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        db.wipe().await.unwrap();
    }

    #[tokio::test]
    async fn query() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello Filters", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert event
        let result = db.save_event(&event).await.unwrap();
        assert!(result);

        // Get event vec from filters
        let filter = Filter::new().author(event.pubkey);
        let filters = vec![filter];
        let result = db.query(filters, Order::Desc).await.unwrap();

        assert!(!result.is_empty());
        assert_eq!(result[0].pubkey, event.pubkey);
    }

    #[tokio::test]
    async fn count() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event_0 = EventBuilder::new(nostr::Kind::TextNote, "Hello Count", vec![])
            .to_event(&keys)
            .unwrap();
        let event_1 = EventBuilder::new(nostr::Kind::TextNote, "Goodbye Count", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert events
        let result = db.save_event(&event_0).await.unwrap();
        assert!(result);
        let result = db.save_event(&event_1).await.unwrap();
        assert!(result);

        // Get event vec from filters
        let filter = Filter::new().author(event_0.pubkey);
        let filters = vec![filter];
        let result = db.count(filters).await.unwrap();

        assert_eq!(result, 2);
    }

    #[tokio::test]
    async fn event_ids_by_filter() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event_0 = EventBuilder::new(nostr::Kind::TextNote, "Hello Count", vec![])
            .to_event(&keys)
            .unwrap();
        let event_1 = EventBuilder::new(nostr::Kind::TextNote, "Goodbye Count", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert events
        let result = db.save_event(&event_0).await.unwrap();
        assert!(result);
        let result = db.save_event(&event_1).await.unwrap();
        assert!(result);

        // Get event vec from filters
        let filter = Filter::new().author(event_0.pubkey);
        let filters = vec![filter];
        let result = db.event_ids_by_filters(filters, Order::Desc).await.unwrap();

        assert_eq!(result.len(), 2);
    }

    #[tokio::test]
    async fn save_parametrized_replaceable_event() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let d_tag = nostr::Tag::Identifier("test".to_string());
        let event = nostr::EventBuilder::new(nostr::Kind::CategorizedPeopleList, "", vec![d_tag])
            .to_event(&keys)
            .unwrap();

        let result = db.save_event(&event).await.unwrap();

        let filter_0 = nostr::Filter::new()
            .kinds(vec![nostr::Kind::ContactList])
            .authors(vec![event.pubkey]);
        let filter_1 = nostr::Filter::new()
            .kinds(vec![nostr::Kind::MuteList])
            .authors(vec![event.pubkey]);
        let filter_2 = nostr::Filter::new()
            .kinds(vec![nostr::Kind::CategorizedPeopleList])
            .authors(vec![event.pubkey])
            .identifiers(vec!["test".to_string()]);
        let filter_3 = nostr::Filter::new()
            .kinds(vec![nostr::Kind::ApplicationSpecificData])
            .authors(vec![event.pubkey])
            .identifiers(vec!["app-config".to_string()]);
        let result = db
            .query(vec![filter_0, filter_1, filter_2, filter_3], Order::Desc)
            .await
            .unwrap();

        assert_eq!(result.len(), 1)
    }

    #[tokio::test]
    async fn save_metadata_event() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;
        let secret_key = "nsec1g24e83hwj5gxl0hqxx9wllwcg9rrxthssv0mrxf4dv3lt8dc29yqrxf09p";
        let keys = nostr::Keys::from_sk_str(secret_key).unwrap();
        // Insert
        let metadata = nostr::Metadata::new()
            .name("Chuck")
            .custom_field("feed", "seed");
        let event = nostr::EventBuilder::metadata(&metadata)
            .to_event(&keys)
            .unwrap();

        let result = db.save_event(&event).await.unwrap();

        // Find profile by pk
        let pubkey = keys.public_key();
        let res = db.get_profile(pubkey).await.unwrap();

        assert_eq!(res.name(), "Sneed");
    }

    #[tokio::test]
    async fn save_event_with_a_tag() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let d_tag = nostr::Tag::Identifier("test".to_string());
        let coordinate_event =
            nostr::EventBuilder::new(nostr::Kind::TextNote, "hello", vec![d_tag.clone()])
                .to_event(&keys)
                .unwrap();
        db.save_event(&coordinate_event).await.unwrap();

        let coordinate = nostr::nips::nip01::Coordinate::new(
            nostr::Kind::CategorizedBookmarkList,
            coordinate_event.pubkey,
        )
        .identifier("test");

        let a_tag = nostr::Tag::A {
            kind: nostr::Kind::CategorizedBookmarkList,
            public_key: keys.public_key(),
            identifier: "test".to_string(),
            relay_url: None,
        };
        let event =
            nostr::EventBuilder::new(nostr::Kind::CategorizedBookmarkList, "", vec![d_tag, a_tag])
                .to_event(&keys)
                .unwrap();

        let result = db.save_event(&event).await.unwrap();

        let ts = nostr::Timestamp::now();
        let res = db
            .has_coordinate_been_deleted(&coordinate, ts)
            .await
            .unwrap();

        dbg!(res);
    }

    // #[tokio::test]
    // async fn get_nip05() {
    //     let config = Arc::new(ServiceConfig::new());
    //     let db = NostrSqlite::new(config).await;

    //     let res = db.get_nip05("test".to_string()).await.unwrap();

    //     dbg!(serde_json::to_value(res).unwrap().to_string());
    // }
}
