use std::collections::HashMap;

use nostr::{key::FromPkStr, Event, Metadata};
use nostr_database::Profile;
use rusqlite::Row;

#[derive(Debug, Clone)]
pub struct EventRow {
    pub id: String,
    pub pubkey: String,
    pub created_at: i64,
    pub kind: i64,
    pub tags: String,
    pub sig: String,
    pub content: String,
}

impl From<&Row<'_>> for EventRow {
    fn from(row: &Row) -> Self {
        let id: String = row.get("id").unwrap();
        let pubkey: String = row.get("pubkey").unwrap();
        let created_at: i64 = row.get("created_at").unwrap();
        let kind: i64 = row.get("kind").unwrap();
        let tags: String = row.get("tags").unwrap();
        let sig: String = row.get("sig").unwrap();
        let content: String = row.get("content").unwrap();

        Self {
            id,
            pubkey,
            created_at,
            kind,
            tags,
            sig,
            content,
        }
    }
}

impl From<&EventRow> for Event {
    fn from(row: &EventRow) -> Self {
        row.to_event()
    }
}

impl EventRow {
    pub fn to_event(&self) -> Event {
        let tags: Vec<Vec<String>> = serde_json::from_str(&self.tags).unwrap();

        let event_json = serde_json::json!(
            {
                "id": self.id,
                "content": self.content,
                "created_at": self.created_at,
                "kind": self.kind,
                "pubkey": self.pubkey,
                "sig": self.sig,
                "tags": tags
            }
        );

        Event::from_value(event_json).unwrap()
    }
}

pub enum EventsTable {
    Table,
    EventId,
    Kind,
    Pubkey,
    Content,
    CreatedAt,
    Tags,
    Sig,
}

impl sea_query::Iden for EventsTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "events",
                Self::EventId => "id",
                Self::Kind => "kind",
                Self::Pubkey => "pubkey",
                Self::Content => "content",
                Self::CreatedAt => "created_at",
                Self::Tags => "tags",
                Self::Sig => "sig",
            }
        )
        .unwrap()
    }
}

pub enum EventsFTSTable {
    Table,
    EventId,
    Content,
}

impl sea_query::Iden for EventsFTSTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "events_fts",
                Self::EventId => "id",
                Self::Content => "content",
            }
        )
        .unwrap()
    }
}

pub enum TagsTable {
    Table,
    Tag,
    Value,
    EventId,
}

impl sea_query::Iden for TagsTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "tags",
                Self::Tag => "tag",
                Self::Value => "value",
                Self::EventId => "event_id",
            }
        )
        .unwrap()
    }
}

pub enum EventSeenByRelaysTable {
    Table,
    Id,
    EventId,
    RelayURL,
}

impl sea_query::Iden for EventSeenByRelaysTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "event_seen_by_relays",
                Self::Id => "id",
                Self::EventId => "event_id",
                Self::RelayURL => "relay_url",
            }
        )
        .unwrap()
    }
}

pub enum ProfilesTable {
    Table,
    Pubkey,
}

impl sea_query::Iden for ProfilesTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "profiles",
                Self::Pubkey => "pubkey",
            }
        )
        .unwrap()
    }
}

pub enum ContactsTable {
    Table,
    Profile,
    Contact,
}

impl sea_query::Iden for ContactsTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "contacts",
                Self::Profile => "profile",
                Self::Contact => "contact",
            }
        )
        .unwrap()
    }
}

pub enum MetadataTable {
    Table,
    Pubkey,
    Name,
    DisplayName,
    About,
    Website,
    Picture,
    Banner,
    Nip05,
    Lud06,
    Lud16,
    Custom,
}

impl sea_query::Iden for MetadataTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "metadata",
                Self::Pubkey => "pubkey",
                Self::Name => "name",
                Self::DisplayName => "display_name",
                Self::About => "about",
                Self::Website => "website",
                Self::Picture => "picture",
                Self::Banner => "banner",
                Self::Nip05 => "nip05",
                Self::Lud06 => "lud06",
                Self::Lud16 => "lud16",
                Self::Custom => "custom",
            }
        )
        .unwrap()
    }
}

#[derive(Debug, Clone)]
pub struct ProfileRow {
    pubkey: String,
    // metadata
    name: Option<String>,
    display_name: Option<String>,
    about: Option<String>,
    website: Option<String>,
    picture: Option<String>,
    banner: Option<String>,
    nip05: Option<String>,
    lud06: Option<String>,
    lud16: Option<String>,
    custom: Option<String>,
}

impl From<&Row<'_>> for ProfileRow {
    fn from(row: &Row) -> Self {
        let pubkey: String = row.get("pubkey").unwrap();
        let name: Option<String> = row.get("name").unwrap_or_default();
        let display_name: Option<String> = row.get("display_name").unwrap_or_default();
        let about = row.get("about").unwrap_or_default();
        let website: Option<String> = row.get("website").unwrap_or_default();
        let picture: Option<String> = row.get("picture").unwrap_or_default();
        let banner: Option<String> = row.get("banner").unwrap_or_default();
        let nip05: Option<String> = row.get("nip05").unwrap_or_default();
        let lud06: Option<String> = row.get("lud06").unwrap_or_default();
        let lud16: Option<String> = row.get("lud16").unwrap_or_default();
        let custom: Option<String> = row.get("custom").unwrap_or_default();

        Self {
            pubkey,
            name,
            display_name,
            about,
            website,
            picture,
            banner,
            nip05,
            lud06,
            lud16,
            custom,
        }
    }
}

impl From<&ProfileRow> for Profile {
    fn from(row: &ProfileRow) -> Self {
        let row = row.to_owned();
        // let f = nostr::EventBuilder::metadata( // Why am I creating this methods in Noose? Just store Kind 0 on a relay and think that the user is registered on the relay ffs
        let keys = nostr::Keys::from_pk_str(&row.pubkey).unwrap();

        let custom_fields: HashMap<String, serde_json::Value> = match row.custom {
            Some(fields_str) => {
                let fields: HashMap<String, serde_json::Value> =
                    serde_json::from_str(&fields_str).unwrap_or(HashMap::new());
                fields
            }
            None => {
                let f: HashMap<String, serde_json::Value> = HashMap::new();
                f
            }
        };

        let metadata = Metadata {
            name: row.name,
            display_name: row.display_name,
            about: row.about,
            website: row.website,
            picture: row.picture,
            banner: row.banner,
            nip05: row.nip05,
            lud06: row.lud06,
            lud16: row.lud16,
            custom: custom_fields,
        };

        nostr_database::Profile::new(keys.public_key(), metadata)
    }
}
