use std::{path::PathBuf, str::FromStr};

use nostr::{key::FromPkStr, secp256k1::XOnlyPublicKey};

#[derive(Clone, Debug)]
pub struct Config {
    admin_pubkey: XOnlyPublicKey,
    db_path: PathBuf,
    auth_required: bool,
    relay_url: nostr::UncheckedUrl,
}

impl Default for Config {
    fn default() -> Self {
        panic!("Use Config::new(env_admin_pk)")
    }
}

impl Config {
    pub fn new() -> Self {
        let admin_pubkey = std::env::var("ADMIN_PUBKEY")
            .map(|env_pk| nostr::Keys::from_pk_str(&env_pk))
            .and_then(|result| result.map_err(|err| panic!("{}", err)))
            .unwrap()
            .public_key();

        let db_path = std::env::var("DATABASE_URL").map(PathBuf::from).unwrap();
        let auth_required: bool = std::env::var("CONFIG_ENABLE_AUTH")
            .unwrap_or("false".to_string())
            .parse()
            .unwrap();
        let relay_url = nostr::UncheckedUrl::from_str(&std::env::var("CONFIG_RELAY_URL").unwrap()).unwrap();

        Self {
            admin_pubkey,
            db_path,
            auth_required,
            relay_url,
        }
    }

    pub fn auth_required(&self) -> bool {
        self.auth_required
    }

    pub fn get_admin_pubkey(&self) -> &XOnlyPublicKey {
        &self.admin_pubkey
    }

    pub fn get_db_path(&self) -> PathBuf {
        self.db_path.clone()
    }

    pub fn get_relay_url(&self) -> nostr::UncheckedUrl {
        self.relay_url.clone()
    }

    pub fn get_relay_config_json(&self) -> serde_json::Value {
        serde_json::json!({
            "contact": "klink@zhitno.st",
            "name": "zhitno.st",
            "description": "Very *special* nostr relay",
            "supported_nips": [ 1, 2, 4, 9, 11, 12, 15, 16, 20, 22, 28, 33, 40, 42, 45, 50 ],
            "software": "git+https://git.zhitno.st/Klink/sneedstr.git",
            "version": "0.1.1"
        })
    }
}
