use argon2::{
    password_hash::{rand_core::OsRng, PasswordHash, PasswordHasher, PasswordVerifier, SaltString},
    Argon2,
};

pub fn password_hash(password: String) -> String {
    let password = password.as_bytes();
    let salt = SaltString::generate(&mut OsRng);

    // Argon2 with default params (Argon2id v19)
    let argon2 = Argon2::default();

    // Hash password to PHC string ($argon2id$v=19$...)
    let password_hash = argon2.hash_password(password, &salt).unwrap().to_string();

    password_hash
}

pub fn password_verify(password: String, password_hash: String) -> bool {
    // Verify password against PHC string.
    //
    // NOTE: hash params from `parsed_hash` are used instead of what is configured in the
    // `Argon2` instance.
    let password = password.as_bytes();
    let parsed_hash = PasswordHash::new(&password_hash).unwrap();
    Argon2::default()
        .verify_password(password, &parsed_hash)
        .is_ok()
}

#[cfg(test)]
mod tests {
    use super::{password_hash, password_verify};

    #[test]
    fn hash() {
        let pass = "mysecretpassword".to_string();
        let hash = password_hash(pass);

        assert_ne!(hash, "".to_string());
    }

    #[test]
    fn verify() {
        let pass = "mysecretpassword".to_string();
        let hash = password_hash(pass.clone());

        let verification = password_verify(pass, hash);

        assert!(verification);
    }
}
